from fizgrid.grid import Grid
from fizgrid.entities import Entity
from fizgrid.utils import Shape

# from pprint import pp as print

grid = Grid(
    name="test_grid",
    x_size=10,
    y_size=10,
    add_exterior_walls=True,
)

# Add some AMRs to the grid
amr1 = grid.add_entity(
    Entity(
        name="AMR1",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=5,
        y_coord=3,
    )
)

amr2 = grid.add_entity(
    Entity(
        name="AMR2",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=3,
        y_coord=5,
    ),
    time=4,
)

amr3 = grid.add_entity(
    Entity(
        name="AMR3",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=5,
        y_coord=5,
    ),
    time=2,
)

grid.remove_entity(entity=amr3, time=3)

# Add routes to the entities such that they will collide

amr1.add_route(
    waypoints=[
        (5, 7, 1),
    ],
)

# Note: We need to wait for this AMR to be in the grid before it can move
amr2.add_route(
    waypoints=[
        (7, 5, 1),
    ],
    time=5,
)

# Run the sim
grid.simulate()

# Show the history of the entities
# print(amr1.history)
# print(amr2.history)
# print(amr3.history)

# Check if the entities resulted in a collision
try:
    success = (
        amr2.history[-1]["x"] == 7
        and amr2.history[-1]["y"] == 5
        and amr2.history[-1]["c"] == False
    )
except:
    success = False
if success:
    print("test_09.py: passed")
else:
    print("test_09.py: failed")



from fizgrid.utils import ShapeMoverUtils


success = True
try:
    # Test case 1: Basic Movement
    seg_start = 0
    seg_end = 1
    t_start = 0
    t_end = 2
    shift = 2

    expected_result = {
        0: (0.0, 1.0),
        1: (0.0, 2.0),
        2: (1.0, 2.0),
    }

    result = ShapeMoverUtils.moving_segment_overlap_intervals(
        seg_start, seg_end, t_start, t_end, shift
    )

    if result != expected_result:
        print("Basic movement test failed")
        success = False
except:
    print("Basic movement test exception")
    success = False

try:
    # Test case 2: No Movement
    seg_start = 0
    seg_end = 1
    t_start = 0
    t_end = 2
    shift = 0

    expected_result = {
        0: (0.0, 2.0),
    }

    result = ShapeMoverUtils.moving_segment_overlap_intervals(
        seg_start, seg_end, t_start, t_end, shift
    )

    if result != expected_result:
        print("No movement test failed")
        success = False
except:
    print("No movement test exception")
    success = False

try:
    # Negative Movement
    seg_start = 0
    seg_end = 1
    t_start = 0
    t_end = 2
    shift = -2
    expected_result = {
        0: (0.0, 1.0),
        -1: (0.0, 2.0),
        -2: (1.0, 2.0),
    }
    result = ShapeMoverUtils.moving_segment_overlap_intervals(
        seg_start, seg_end, t_start, t_end, shift
    )
    if result != expected_result:
        print("Negative movement test failed")
        success = False
except:
    print("Negative movement test exception")
    success = False

try:
    # Test case 4: Zero Duration
    seg_start = 0
    seg_end = 1
    t_start = 0
    t_end = 0
    shift = 2

    expected_result = {}

    result = ShapeMoverUtils.moving_segment_overlap_intervals(
        seg_start, seg_end, t_start, t_end, shift
    )

    if result != expected_result:
        print("Zero duration test failed")
        success = False
except:
    print("Zero duration test exception")
    success = False


try:
    # Test with a rectangle
    x_start = 0
    x_end = 1
    y_start = 0
    y_end = 1
    x_shift = 2
    y_shift = 2
    t_start = 0
    t_end = 2
    expected_result = {
        (0, 0): (0.0, 1.0),
        (0, 1): (0.0, 1.0),
        (1, 0): (0.0, 1.0),
        (1, 1): (0.0, 2.0),
        (2, 1): (1.0, 2.0),
        (1, 2): (1.0, 2.0),
        (2, 2): (1.0, 2.0),
    }

    result = ShapeMoverUtils.moving_rectangle_overlap_intervals(
        x_start=x_start,
        x_end=x_end,
        y_start=y_start,
        y_end=y_end,
        x_shift=x_shift,
        y_shift=y_shift,
        t_start=t_start,
        t_end=t_end,
    )

    if expected_result != result:
        print("Rectangle test failed")
        success = False
except:
    print("Rectangle test exception")
    success = False

try:
    # test with a triangle moving from (0,0) to (2,2)
    # Use a triangle at (0,0), (1,0), (0,1)
    shape = [[0, 0], [1, 0], [0, 1]]
    x_coord = 0
    y_coord = 0
    x_shift = 2
    y_shift = 2
    t_start = 0
    t_end = 2
    expected_result = {
        (0, 0): (0.0, 1.0),
        (0, 1): (0.0, 1.0),
        (1, 0): (0.0, 1.0),
        (1, 1): (0.0, 2.0),
        (2, 1): (1.0, 2.0),
        (1, 2): (1.0, 2.0),
        (2, 2): (1.0, 2.0),
    }

    result = ShapeMoverUtils.moving_shape_overlap_intervals(
        shape=shape,
        x_coord=x_coord,
        y_coord=y_coord,
        x_shift=x_shift,
        y_shift=y_shift,
        t_start=t_start,
        t_end=t_end,
    )
    if expected_result != result:
        print("Shape test positive slope failed")
        success = False
except:
    print("Shape test positive slope exception")
    success = False

try:
    # test with a triangle moving from 0,0 down to 1,-1
    # Use a triangle at (0,0), (1,0), (0,1)
    shape = [[0, 0], [1, 0], [0, 1]]
    x_coord = 0
    y_coord = 0
    x_shift = 1
    y_shift = -1
    t_start = 0
    t_end = 2
    expected_result = {(0, -1): (0, 2.0), (0, 0): (0, 2.0), (1, -1): (0.0, 2)}
    result = ShapeMoverUtils.moving_shape_overlap_intervals(
        shape=shape,
        x_coord=x_coord,
        y_coord=y_coord,
        x_shift=x_shift,
        y_shift=y_shift,
        t_start=t_start,
        t_end=t_end,
    )
    if expected_result != result:
        print("Shape test negative slope with removals failed")
        success = False
except:
    print("Shape test negative slope with removals exception")
    success = False

try:
    # test with a triangle moving from 0,0 up to 1,1
    # Use a triangle at (0,0), (1,0), (1,1)
    shape = [[0, 0], [1, 0], [1, 1]]
    x_coord = 0
    y_coord = 0
    x_shift = 1
    y_shift = 1
    t_start = 0
    t_end = 2
    expected_result = {(0, 0): (0, 2.0), (1, 0): (0.0, 2), (1, 1): (0.0, 2)}
    result = ShapeMoverUtils.moving_shape_overlap_intervals(
        shape=shape,
        x_coord=x_coord,
        y_coord=y_coord,
        x_shift=x_shift,
        y_shift=y_shift,
        t_start=t_start,
        t_end=t_end,
    )
    if expected_result != result:
        print("Shape test positive slope with removals failed")
        success = False
except:
    print("Shape test positive slope with removals exception")
    success = False

if success:
    print("test_03.py: passed")
else:
    print("test_03.py: failed")



from fizgrid.grid import Grid
from fizgrid.entities import Entity
from fizgrid.utils import Shape

# from pprint import pp as print

grid = Grid(
    name="test_grid",
    x_size=10,
    y_size=10,
    add_exterior_walls=True,
    cell_density=2,
)

# Add some AMRs to the grid
amr1 = grid.add_entity(
    Entity(
        name="AMR1",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=5,
        y_coord=3,
    )
)
amr2 = grid.add_entity(
    Entity(
        name="AMR2",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=3,
        y_coord=5,
    )
)

# Add routes to the entities such that they will collide

amr1.add_route(
    waypoints=[
        (5, 7, 1),
    ],
)

amr2.add_route(
    waypoints=[
        (7, 5, 1),
    ],
)

# Run the sim
next_state_events = True
while next_state_events:
    next_state_events = grid.resolve_next_state()
# Check if the entities resulted in a collision
try:
    success = amr1.history[-1]["y"] == 4 and amr1.history[-1]["c"] == True
except:
    success = False
if success:
    print("test_04.py: passed")
else:
    print("test_04.py: failed")



def simple_waypoint_time_approximation(
    start_x: int | float,
    start_y: int | float,
    waypoints: list[tuple],
    speed: int | float,
):
    """
    A helper function to approximate waypoint times assuming a constant entity speed
    Returns the waypoints with the time it would take to travel between them given the constant speed.

    Args:

    - start_x: The x coordinate of the starting point.
    - start_y: The y coordinate of the starting point.
    - waypoints: A list of tuples representing the waypoints to be traversed by the moving entity.
    - speed: The constant speed of the moving entity.

    Returns:

    - A list of tuples representing the waypoints with the time it would take to travel between them given the constant speed.
        - Each tuple contains the x and y coordinates of the original waypoint as well as the time delta to travel to said waypoint.
    """
    # Store output to return
    output = []
    tmp_x = start_x
    tmp_y = start_y
    for waypoint in waypoints:
        distance_traveled = (
            (waypoint[0] - tmp_x) ** 2 + (waypoint[1] - tmp_y) ** 2
        ) ** 0.5
        time_to_travel = distance_traveled / speed
        tmp_x = waypoint[0]
        tmp_y = waypoint[1]
        output.append((tmp_x, tmp_y, time_to_travel))

    return output



import math


def get_distance(pt1: tuple, pt2: tuple):
    """
    Returns the Euclidean distance between two points.

    Args:

    - pt1: The first point (x, y).
    - pt2: The second point (x, y).

    Returns:

    - float: The distance between the two points.
    """
    return ((pt1[0] - pt2[0]) ** 2 + (pt1[1] - pt2[1]) ** 2) ** 0.5


def get_angle(pt1: tuple, pt2: tuple, pt3: tuple):
    """
    Calculates the angle when going from pt1 to pt2 to pt3.

    Returns the angle in degrees.

    Args:

    - pt1: The first point (x, y).
    - pt2: The second point (x, y).
    - pt3: The third point (x, y).

    Returns:

    - float: The angle in degrees.

    Examples:

    EG: (0, 0) -> (1, 0) -> (2, 0) => 180
    EG: (0, 0) -> (1, 0) -> (1, 1) => 90
    """
    v1 = (pt1[0] - pt2[0], pt1[1] - pt2[1])
    v2 = (pt3[0] - pt2[0], pt3[1] - pt2[1])
    if v1 == (0, 0) or v2 == (0, 0):
        return 0.0
    magnitude_v1 = (v1[0] ** 2 + v1[1] ** 2) ** 0.5
    magnitude_v2 = (v2[0] ** 2 + v2[1] ** 2) ** 0.5
    dot_product = v1[0] * v2[0] + v1[1] * v2[1]
    cos_theta = dot_product / (magnitude_v1 * magnitude_v2)
    cos_theta_clamp = max(-1, min(1, (cos_theta)))
    return math.acos(cos_theta_clamp) / math.pi * 180


def get_acceleration_distance(
    start_speed: int | float, end_speed: int | float, acceleration: int | float
):
    """
    Calculates the distance required to accelerate/decelerate from start_speed to end_speed
    given a constant acceleration.

    Args:

    - start_speed: The speed at which the object is currently traveling.
    - end_speed: The speed at which the object will be traveling.
    - acceleration: The acceleration of the object (should always be positive as start and end speeds determine if it is accelerating or decelerating).

    Returns:

    - float: The distance required to accelerate/decelerate from start_speed to end_speed.
        - If start_speed and end_speed are equal, 0 is returned.
        - If start_speed is greater than end_speed, the function will return the distance required to decelerate from start_speed to end_speed.
        - If start_speed is less than end_speed, the function will return the distance required to accelerate from start_speed to end_speed.
    """
    if start_speed == end_speed:
        return 0.0
    if start_speed > end_speed:
        start_speed, end_speed = end_speed, start_speed
    return (end_speed**2 - start_speed**2) / (2 * abs(acceleration))


def get_end_speed(
    start_speed: int | float, distance: int | float, acceleration: int | float
):
    """
    Calculates the speed at wich an object will be traveling after a given distance
    given a constant acceleration.

    Args:

    - start_speed: The speed at which the object is currently traveling.
    - distance: The distance over which the object will be traveling.
    - acceleration: The acceleration/deceleration of the object (positive for acceleration, negative for deceleration).

    Returns:

    - float: The speed at which the object will be traveling after the given distance.
    """
    if acceleration == 0 or distance == 0:
        return start_speed
    return (start_speed**2 + 2 * acceleration * distance) ** 0.5


def partition_distance(
    distance: int | float,
    start_speed: int | float,
    max_end_speed: int | float,
    max_speed: int | float,
    acceleration: int | float,
):
    """
    Partitions a distance into time and distance for acceleration, cruise and deceleration

    args:

    - distance: The distance to be traveled.
    - start_speed: The speed at which the entity is currently traveling.
    - max_end_speed: The maximum ending speed that the entity can travel at.
    - max_speed: The maximum speed of the entity.
    - acceleration: The acceleration/deceleration of the entity.

    Returns a tuple
    - 0: list of dictionaries with the following keys:
        - end_pct: The percentage of the distance at which the partition ends.
        - time: The time it takes to travel the distance in the partition.
    - 1: int or float: The ending speed

    """
    # Compute peak speed possible within distance
    max_speed_possible_in_distance = get_end_speed(
        start_speed=start_speed, distance=distance, acceleration=acceleration
    )
    max_end_speed = min(max_end_speed, max_speed_possible_in_distance)
    peak_speed = min(
        ((2 * acceleration * distance + start_speed**2 + max_end_speed**2) / 2)
        ** 0.5,
        max_speed,
    )

    end_speed = min(max_end_speed, peak_speed)
    partitions = []
    accel_end_pct = 0
    cruise_end_pct = 1

    # Compute distances and times for acceleration, cruise and deceleration
    if peak_speed > start_speed:
        accel_time = (peak_speed - start_speed) / acceleration
        accel_dist = get_acceleration_distance(
            start_speed, peak_speed, acceleration
        )
        accel_end_pct = accel_dist / distance
        partitions.append({"end_pct": accel_end_pct, "time": accel_time})
    if peak_speed > max_end_speed:
        decel_time = (peak_speed - max_end_speed) / acceleration
        decel_dist = get_acceleration_distance(
            peak_speed, max_end_speed, acceleration
        )
        cruise_end_pct = (distance - decel_dist) / distance
        partitions.append({"end_pct": 1, "time": decel_time})
    if peak_speed == max_speed:
        cruise_dist = (cruise_end_pct - accel_end_pct) * distance
        cruise_time = cruise_dist / peak_speed
        partitions.append({"end_pct": cruise_end_pct, "time": cruise_time})
    return sorted(partitions, key=lambda x: x["end_pct"]), end_speed



from fizgrid.grid import Grid
from fizgrid.entities import Entity, StaticEntity

# This test is special. It creates rounding errors in python that create a permanent collision
# This is fixed in the code using `location_precision` for each entity to fix any rounding errors before they cause issues.

# from pprint import pp as print
grid = Grid(
    name="test_grid",
    x_size=10,
    y_size=20,
    add_exterior_walls=True,
)

amr = grid.add_entity(
    Entity(
        name="AMR2",
        shape=[[-1.5, -0.5], [-1.5, 0.5], [1.5, 0.5], [1.5, -0.5]],
        x_coord=5,
        y_coord=5,
    )
)

wall = grid.add_entity(
    StaticEntity(
        name="Wall1",
        shape=[[8, 0], [0, 0], [0, 1], [8, 1]],
        x_coord=1,
        y_coord=10,
    )
)

amr.add_route(
    waypoints=[
        (5, 15, 1),
    ],
    time=5,
)

# Run the sim
grid.simulate()

# Check if a collision is loggged
success = True
try:
    if not amr.history[-1]["c"]:
        success = False
except:
    success = False
if success:
    print("test_12.py: passed")
else:
    print("test_12.py: failed")



import type_enforced, heapq


@type_enforced.Enforcer(enabled=True)
class TimeQueue:
    def __init__(self):
        """
        Initializes a TimeQueue instance.
        This class is used to manage a queue of events that occur at specific times.
        It uses a min-heap to efficiently manage the events based on their scheduled times.
        """
        self.__heap__ = []
        self.__data__ = {}
        self.__time__ = 0
        self.__next_id__ = 0

    def add_event(
        self, time: int | float, event: dict = dict(), priority: int = 0
    ) -> int:
        """
        Adds an event to the queue.

        Args:

        - time (int|float): The time at which the event should occur.
        - event (dict): The event to be added to the queue.
            - Default: {}
            - This have any dictionary strucutre, depending on your queue needs
        - priority (int): The priority of the event.
            - Default: 0
            - Higher values indicate higher priority.
            - This is used to determine the order of events with the same time.
            - If two events have the same time, the one with the higher priority will be processed first.
            - If the priority is the same, the event with the lower ID will be processed first.

        Returns:

        - int: The ID of the added event.
            - This ID is used to reference the event in the queue.
        """
        assert (
            time >= self.__time__
        ), "Time must be greater than or equal to current time"
        id = self.__next_id__
        self.__next_id__ += 1
        self.__data__[id] = event
        heapq.heappush(self.__heap__, (time, -priority, id))
        return id

    def remove_event(self, id: int) -> dict | None:
        """
        Removes an event from the queue using its ID.

        Args:

            - id (int): The ID of the event to be removed.
                - This ID is used to reference the event in the queue.
        Returns:
            - dict: The removed event.
                - If the event is not found, None is returned.
        """
        return self.__data__.pop(id, None)

    def remove_next_event(self) -> dict | None:
        """
        Removes the next event from the queue.
        This method is used to get the next event in the queue and remove it from the heap.

        Returns:

        - dict: The removed event.
            - If the queue is empty, None is returned.
        """
        self.remove_event(heapq.heappop(self.__heap__)[2])

    def get_next_event(self, peek: bool = False):
        """
        Retrieves the next event from the queue without removing it.
        This method is used to get the next event in the queue

        Args:

        - peek (bool): If True, the event is not removed from the queue.
            - Default: False
            - If False, the event is removed from the queue.

        Returns:

        - dict: The next event in the queue.
            - If the queue is empty, None is returned.
        """
        while self.__heap__:
            if peek:
                time, priority, id = self.__heap__[0]
                event = self.__data__.get(id, None)
                if event is None:
                    # Remove the event from the heap to avoid stale references
                    heapq.heappop(self.__heap__)
                    continue
            else:
                time, priority, id = heapq.heappop(self.__heap__)
                event = self.remove_event(id)
                if event is None:
                    continue
                self.__time__ = time
            return {
                "id": id,
                "time": time,
                "event": event,
            }
        return {
            "id": None,
            "time": None,
            "event": None,
        }

    def get_next_events(self) -> list[dict]:
        """
        Retrieves all events that occur at the same time as the next event.
        This method is used to get all events that occur at the same time as the next event in the queue.
        It removes these events from the queue.

        Returns:

        - list: A list of events that occur at the same time as the next event.
            - If the queue is empty, an empty list is returned.
        """
        events = []
        event = self.get_next_event(peek=True)
        if event["time"] != None:
            self.__time__ = event["time"]
            next_event = event
            while next_event["time"] == self.__time__:
                event = next_event
                self.remove_next_event()
                events.append(event)
                next_event = self.get_next_event(peek=True)
        return events



# Testing for TimeQueue class
from fizgrid.helpers.waypoint_timing.simple import (
    simple_waypoint_time_approximation,
)
from fizgrid.helpers.waypoint_timing.acceleration import (
    acceleration_waypoint_time_approximation,
)

passing = True

test_case_1 = [(1, 0), (2, 0), (2, 1)]

simple_timed_waypoints = simple_waypoint_time_approximation(
    start_x=0, start_y=0, waypoints=test_case_1, speed=1
)

if simple_timed_waypoints != [(1, 0, 1.0), (2, 0, 1.0), (2, 1, 1.0)]:
    print("test_13.py: simple_waypoint_time_approximation failed")
    passing = False

acceleration_timed_waypoints = acceleration_waypoint_time_approximation(
    start_x=0, start_y=0, waypoints=test_case_1, max_speed=1, acceleration=1
)

if acceleration_timed_waypoints != [
    (0.5, 0.0, 1.0),
    (1, 0, 0.5),
    (1.505, 0.0, 0.505),
    (2, 0, 0.9),
    (2.0, 0.495, 0.9),
    (2.0, 0.5, 0.005),
    (2, 1, 1.0),
]:
    print("test_13.py: acceleration_waypoint_time_approximation failed")
    passing = False


if passing:
    print("test_13.py: passed")
else:
    print("test_13.py: failed")



import math, type_enforced, uuid


def unique_id():
    """
    Generates a unique identifier.
    """
    return str(uuid.uuid4())


@type_enforced.Enforcer(enabled=True)
class Shape:
    @staticmethod
    def circle(
        radius: int | float, num_points: int = 6, round_to: int = 2
    ) -> list[list]:
        """
        Returns a list of addative coordinates that form a circle around a given point.

        Args:

        - radius (int|float): Radius of the circle.
        - num_points (int): Number of points to generate around the circle.
            - Default: 6
            - This is used to determine the number of points to generate around the circle.
        - round_to (int): Number of decimal places to round to.
            - Default: 2
            - This is used to round the coordinates to a specific number of decimal places.

        Returns:

        - list[list]: A list of coordinates representing the circle.
            - Each coordinate is a list of two values [x, y].
            - The coordinates are relative to the center of the circle.
        """
        return [
            [
                round(
                    radius * math.cos(2 * math.pi / num_points * i), round_to
                ),
                round(
                    radius * math.sin(2 * math.pi / num_points * i), round_to
                ),
            ]
            for i in range(num_points)
        ]

    @staticmethod
    def rectangle(
        x_len: float | int, y_len: float | int, round_to: int = 2
    ) -> list[list]:
        """
        Returns a list of addative coordinates that form a rectangle around a given point.

        Args:

        - x_len (float|int): Length of the rectangle along the x-axis.
        - y_len (float|int): Length of the rectangle along the y-axis.
        - round_to (int): Number of decimal places to round to.
            - Default: 2
            - This is used to round the coordinates to a specific number of decimal places.

        Returns:

        - list[list]: A list of coordinates representing the rectangle.
            - Each coordinate is a list of two values [x, y].
            - The coordinates are relative to the center of the rectangle.
        """
        return [
            [round(x_len / 2, round_to), round(y_len / 2, round_to)],
            [round(-x_len / 2, round_to), round(y_len / 2, round_to)],
            [round(-x_len / 2, round_to), round(-y_len / 2, round_to)],
            [round(x_len / 2, round_to), round(-y_len / 2, round_to)],
        ]

    @staticmethod
    def rotate(
        radians: float | int, shape: list[list[float | int]], round_to: int = 2
    ) -> list[list[float | int]]:
        """
        Rotates a shape by a given angle in radians around its origin (0, 0).
        Args:
        - radians (float|int): The angle in radians to rotate the shape.
        - shape (list[list[float|int]]): The shape to rotate.
            - This is a list of coordinates representing the shape.
            - Each coordinate is a list of two values [x, y].
        Returns:
        - list[list[float|int]]: The rotated shape.
            - This is a list of coordinates representing the rotated shape.
            - Each coordinate is a list of two values [x, y].
        """
        return [
            [
                round(
                    coord[0] * math.cos(radians) - coord[1] * math.sin(radians),
                    round_to,
                ),
                round(
                    coord[0] * math.sin(radians) + coord[1] * math.cos(radians),
                    round_to,
                ),
            ]
            for coord in shape
        ]

    @staticmethod
    def get_rotated_shape(
        shape: list[list[float | int]],
        x_shift: float | int,
        y_shift: float | int,
    ):
        """
        Rotates the shape to align with the direction of motion.

        Args:

        - shape (list[list[float|int]]): List of coordinates representing the shape's vertices relative to its center.
        - x_shift (float|int): Total distance the shape moves along the x-axis during [t_start, t_end].
        - y_shift (float|int): Total distance the shape moves along the y-axis during [t_start, t_end].

        Returns:

        - list[list[float|int]]: The rotated shape.
            - This is a list of coordinates representing the rotated shape.
            - Each coordinate is a list of two values [x, y].
        """
        if x_shift == 0 and y_shift == 0:
            return shape
        radians = math.atan2(y_shift, x_shift)
        return Shape.rotate(radians=radians, shape=shape)


class ShapeMoverUtils:
    @staticmethod
    def moving_segment_overlap_intervals(
        seg_start: int | float,
        seg_end: int | float,
        t_start: int | float,
        t_end: int | float,
        shift: int | float,
    ):
        """
        Calculates the time intervals during which a moving 1D line segment overlaps with each unit-length
        integer-aligned range along the x-axis.

        Args:

        - seg_start (int|float): Initial position of the left end of the line segment.
        - seg_end (int|float): Initial position of the right end of the line segment.
        - t_start (int|float): Start time of the motion.
        - t_end (int|float): End time of the motion.
        - shift (int|float): Total distance the line segment moves along the x-axis during [t_start, t_end].

        Returns:

        - dict[int, tuplie(int|float,int|float)]: A dictionary mapping each integer `i` to the time interval [t_in, t_out]
                                during which any part of the line overlaps the range [i, i+1).
                                Only includes ranges with non-zero overlap duration.
        """
        duration = t_end - t_start
        velocity = shift / duration if duration != 0 else 0

        result = {}

        final_start = seg_start + shift
        final_end = seg_end + shift
        global_min = min(seg_start, final_start)
        global_max = max(seg_end, final_end)

        for i in range(int(global_min) - 1, int(global_max) + 2):
            if velocity == 0:
                if seg_end > i and seg_start < i + 1 and t_start < t_end:
                    result[i] = (t_start, t_end)
                continue
            # Solve for times when the line enters and exits overlap with [i, i+1)
            t1 = (i - seg_end) / velocity + t_start
            t2 = (i + 1 - seg_start) / velocity + t_start
            entry_time = max(min(t1, t2), t_start)
            exit_time = min(max(t1, t2), t_end)

            if exit_time > entry_time:
                result[i] = (entry_time, exit_time)

        return result

    @staticmethod
    def moving_rectangle_overlap_intervals(
        x_start: float | int,
        x_end: float | int,
        y_start: float | int,
        y_end: float | int,
        x_shift: float | int,
        y_shift: float | int,
        t_start: float | int,
        t_end: float | int,
    ):
        """
        Calculates the time intervals during which a moving rectangle overlaps with each unit-length
        integer-aligned range along the x and y axes.

        Args:

        - x_start (float|int): Initial position of the left end of the rectangle along the x-axis.
        - x_end (float|int): Initial position of the right end of the rectangle along the x-axis.
        - y_start (float|int): Initial position of the bottom end of the rectangle along the y-axis.
        - y_end (float|int): Initial position of the top end of the rectangle along the y-axis.
        - x_shift (float|int): Total distance the rectangle moves along the x-axis during [t_start, t_end].
        - y_shift (float|int): Total distance the rectangle moves along the y-axis during [t_start, t_end].
        - t_start (float|int): Start time of the motion.
        - t_end (float|int): End time of the motion.

        Returns:

        - dict[tuple(int,int),tuple(int|float,int|float)]: A dictionary mapping each integer (i,j) to the time interval [t_in, t_out]
            during which any part of the rectangle overlaps the range [i, i+1) x [j, j+1).
            Only includes ranges with non-zero overlap duration.

        """
        x_intervals = ShapeMoverUtils.moving_segment_overlap_intervals(
            seg_start=x_start,
            seg_end=x_end,
            t_start=t_start,
            t_end=t_end,
            shift=x_shift,
        )
        y_intervals = ShapeMoverUtils.moving_segment_overlap_intervals(
            seg_start=y_start,
            seg_end=y_end,
            t_start=t_start,
            t_end=t_end,
            shift=y_shift,
        )
        result = {}
        for x_key, x_interval in x_intervals.items():
            for y_key, y_interval in y_intervals.items():
                # Only add intervals with time overlap
                if (
                    x_interval[1] > y_interval[0]
                    and y_interval[1] > x_interval[0]
                ):
                    result[(x_key, y_key)] = (
                        max(x_interval[0], y_interval[0]),
                        min(x_interval[1], y_interval[1]),
                    )

        return result

    @staticmethod
    def argmin(lst):
        return min(enumerate(lst), key=lambda x: x[1])[0]

    @staticmethod
    def argmax(lst):
        return max(enumerate(lst), key=lambda x: x[1])[0]

    @staticmethod
    def find_extreme_orthogonal_vertices(
        points: list[tuple[int | float, int | float]], slope: float | int
    ):
        """
        Finds the points in a list that are the furthest apart in the direction
        orthogonal to the given slope. This is useful for finding the extreme
        points in a set of 2D coordinates that are orthogonal to a given line.

        Args:

        - points (list of tuples): A list of (x, y) coordinates.
        - slope (float): The slope of the line.

        Returns:

        - tuple: The points with the minimum and maximum projections.
        """
        # Compute orthogonal slope
        orthogonal = float("inf") if slope == 0 else -1 / slope
        # Define direction projections (a normalized direction vector, but without the linear algebra)
        # Handle vertical slope
        if orthogonal == float("inf"):
            x_proj = 0.0
            y_proj = 1.0
        else:
            # Note: Technically normalized length is (1**2 + orthogonal**2)**.5, but we avoid the extra square for performance
            length = (1 + orthogonal**2) ** 0.5
            x_proj = 1.0 / length
            y_proj = orthogonal / length
        # Compute projections
        projections = [x * x_proj + y * y_proj for x, y in points]
        # Return the min and max projections
        return (
            points[ShapeMoverUtils.argmin(projections)],
            points[ShapeMoverUtils.argmax(projections)],
        )

    @staticmethod
    def find_extreme_orthogonal_vertices_simplified(
        points: list[tuple[int | float, int | float]], slope: float | int
    ):
        """
        A simplified version of the function `find_extreme_orthogonal_vertices`
        that assumes the slope is never 0.

        Finds the points in a list that are the furthest apart in the direction
        orthogonal to the given slope. This is useful for finding the extreme
        points in a set of 2D coordinates that are orthogonal to a given line.

        Args:

        - points (list of tuples): A list of (x, y) coordinates.
        - slope (float): The slope of the line.
            - Note: This should never be 0 or infinite for this function.

        Returns:

        - tuple: The points with the minimum and maximum projections.
        """
        # Compute orthogonal slope
        orthogonal = -1 / slope
        # Define direction projections (a normalized direction vector, but without the linear algebra)
        # Note: Technically normalized length is (1**2 + orthogonal**2)**.5, but we avoid the extra square for performance
        length = (1 + orthogonal**2) ** 0.5
        projections = [
            x * 1.0 / length + y * orthogonal / length for x, y in points
        ]
        # Return the min and max verticies
        vertex_1 = points[ShapeMoverUtils.argmin(projections)]
        vertex_2 = points[ShapeMoverUtils.argmax(projections)]
        if slope < 0:
            return vertex_1, vertex_2
        else:
            return vertex_2, vertex_1

    @staticmethod
    def remove_untouched_intervals(
        intervals: dict[tuple[int, int], tuple[int | float, int | float]],
        slope: float | int,
        absolute_shape: list[tuple[int | float, int | float]],
    ):
        """
        Removes unnecessary intervals from the dictionary of intervals.

        Args:

        - intervals (dict[tuple(int,int),tuple(int|float,int|float)]): A dictionary mapping each integer (i,j) to the time interval [t_in, t_out]
            during which any part of the shape overlaps the range [i, i+1) x [j, j+1).
        - slope (float|int): The slope of the line.
        - absolute_shape (list(tuple[int|float, int|float])): A list of coordinates representing the shape's vertices relative to its center.

        Returns:

        - dict[tuple(int,int),tuple(int|float,int|float)]: A dictionary with unnecessary intervals removed.
        """
        min_vertex, max_vertex = (
            ShapeMoverUtils.find_extreme_orthogonal_vertices_simplified(
                points=absolute_shape, slope=slope
            )
        )
        shape_min_intercept = min_vertex[1] - slope * min_vertex[0]
        shape_max_intercept = max_vertex[1] - slope * max_vertex[0]

        # Given the slope, determine which cell vertex to use for the max and min intercepts
        # This is to ensure that the intervals are removed correctly
        ltx_increment = 1 if slope < 0 else 0
        gtx_increment = 0 if slope < 0 else 1

        remove_keys = []
        for x_cell, y_cell in intervals.keys():
            cell_min_intercept = y_cell - (slope * (x_cell + gtx_increment))
            cell_max_intercept = (y_cell + 1) - (
                slope * (x_cell + ltx_increment)
            )
            # If the intercept ranges do not overlap, remove the interval
            if not (
                cell_min_intercept < shape_max_intercept
                and shape_min_intercept < cell_max_intercept
            ):
                remove_keys.append((x_cell, y_cell))
        for key in remove_keys:
            del intervals[key]
        return intervals

    @staticmethod
    def moving_shape_overlap_intervals(
        x_coord: float | int,
        y_coord: float | int,
        x_shift: float | int,
        y_shift: float | int,
        t_start: float | int,
        t_end: float | int,
        shape: list[list[float | int]],
        cell_density: int = 1,
    ):
        """
        Calculates the time intervals during which a moving shape overlaps with each unit-length
        integer-aligned range along the x and y axes.

        Note: This converts each shape into a full bounding box rectangle and then uses the rectangle overlap function to calculate the intervals.

        Args:

        - x_coord (float|int): Initial x-coordinate of the shape's center.
        - y_coord (float|int): Initial y-coordinate of the shape's center.
        - x_shift (float|int): Total distance the shape moves along the x-axis during [t_start, t_end].
        - y_shift (float|int): Total distance the shape moves along the y-axis during [t_start, t_end].
        - t_start (float|int): Start time of the motion.
        - t_end (float|int): End time of the motion.
        - shape (list[list[float|int]]): List of coordinates representing the shape's vertices relative to its center.
        - cell_density (int): The number of cells per unit of length.


        Returns:

        - dict[tuple(int,int),tuple(int|float,int|float)]: A dictionary mapping each integer (i,j) to the time interval [t_in, t_out]
                                during which any part of the shape overlaps the range [i, i+1) x [j, j+1).
                                Only includes ranges with non-zero overlap duration.
        """
        # Get the overlap intervals for a rectangle that bounds the shape in cell space
        absolute_shape = [
            [
                (x_coord + coord[0]) * cell_density,
                (y_coord + coord[1]) * cell_density,
            ]
            for coord in shape
        ]
        # Get the shifts in cell space
        x_shift = x_shift * cell_density
        y_shift = y_shift * cell_density
        # Calculate the rectangle overlap intervals in cell space
        rectangle_overlap_intervals = (
            ShapeMoverUtils.moving_rectangle_overlap_intervals(
                x_start=min([coord[0] for coord in absolute_shape]),
                x_end=max([coord[0] for coord in absolute_shape]),
                y_start=min([coord[1] for coord in absolute_shape]),
                y_end=max([coord[1] for coord in absolute_shape]),
                x_shift=x_shift,
                y_shift=y_shift,
                t_start=t_start,
                t_end=t_end,
            )
        )
        # If the shape is only moving vertically or horizontally, we can just return the rectangle overlap intervals
        if x_shift == 0 or y_shift == 0:
            return rectangle_overlap_intervals
        # If the shape is moving diagonally, we should remove intervals that are never passed through by the shape
        return ShapeMoverUtils.remove_untouched_intervals(
            intervals=rectangle_overlap_intervals,
            slope=y_shift / x_shift,
            absolute_shape=absolute_shape,
        )



from fizgrid.grid import Grid
from fizgrid.entities import Entity
from fizgrid.utils import Shape
import random, math


class AMR(Entity):
    def __init__(self, *args, server, **kwargs):
        super().__init__(*args, **kwargs)
        self.server = server

    def on_realize(self, **kwargs):
        self.server.request_next_step(self.id)


class Order:
    def __init__(
        self,
        time,
        origin_x,
        origin_y,
        destination_x,
        destination_y,
    ):
        self.time = time
        self.origin_x = origin_x
        self.origin_y = origin_y
        self.destination_x = destination_x
        self.destination_y = destination_y
        self.history = []
        self.set_status(time=time, status="arrived")

    def set_status(self, time, status):
        self.time = time
        self.status = status
        self.history.append({"time": time, "status": status})


class AssignedOrderHandler:
    def __init__(self, amr, order, server):
        self.amr = amr
        self.order = order
        self.server = server
        self.goal_x = None
        self.goal_y = None
        self.goal_tolerance = 1
        self.speed = 1

    def increment_status(self):
        if self.order.status == "in_queue":
            self.goal_x = self.order.origin_x
            self.goal_y = self.order.origin_y
            self.order.set_status(self.amr.get_time(), "waiting_pickup")
            self.handle_next_step()
        elif self.order.status == "waiting_pickup":
            self.goal_x = self.order.destination_x
            self.goal_y = self.order.destination_y
            self.order.set_status(self.amr.get_time(), "picked_up")
            self.handle_next_step()
        elif self.order.status == "picked_up":
            self.goal_x = None
            self.goal_y = None
            self.order.set_status(self.amr.get_time(), "delivered")
            self.server.order_handlers.pop(self.amr.id)
            self.server.add_available_amr(self.amr)
            self.server.completed_orders.append(self.order)

    def get_dist_from_goal(self):
        return (
            (self.goal_x - self.amr.x_coord) ** 2
            + (self.goal_y - self.amr.y_coord) ** 2
        ) ** 0.5

    def handle_next_step(self):
        if self.goal_x is None or self.goal_y is None:
            self.increment_status()
            return
        distance_from_goal = self.get_dist_from_goal()
        if distance_from_goal < self.goal_tolerance:
            self.increment_status()
            return
        goal_angle_rad = math.atan2(
            self.goal_y - self.amr.y_coord, self.goal_x - self.amr.x_coord
        )
        random_angle = random.normalvariate(goal_angle_rad, math.pi / 2)
        distance = random.uniform(0, min(distance_from_goal, 5))
        x_shift = distance * math.cos(random_angle)
        y_shift = distance * math.sin(random_angle)
        self.amr.add_route(
            waypoints=[
                (
                    self.amr.x_coord + x_shift,
                    self.amr.y_coord + y_shift,
                    distance / self.speed,
                ),
            ]
        )


class Server:
    def __init__(self, grid):
        self.grid = grid

        self.orders = []
        self.order_queue = []
        self.completed_orders = []

        self.available_amrs = []
        self.order_handlers = {}

    def schedule_order(
        self, time, origin_x, origin_y, destination_x, destination_y
    ):
        self.grid.add_event(
            time=time,
            object=self,
            method="receive_order",
            kwargs={
                "time": time,
                "origin_x": origin_x,
                "origin_y": origin_y,
                "destination_x": destination_x,
                "destination_y": destination_y,
            },
        )

    def create_amr(self, name, x_coord, y_coord):
        amr = grid.add_entity(
            AMR(
                name=name,
                shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
                x_coord=x_coord,
                y_coord=y_coord,
                server=self,
            )
        )
        self.add_available_amr(amr)
        return amr

    def receive_order(self, *args, **kwargs):
        order = Order(*args, **kwargs)
        self.order_queue.append(order)
        self.orders.append(order)
        order.set_status(order.time, "in_queue")
        self.try_assignment()

    def add_available_amr(self, amr):
        self.available_amrs.append(amr)
        self.try_assignment()

    def try_assignment(self):
        if len(self.order_queue) == 0 or len(self.available_amrs) == 0:
            return
        order_handler = AssignedOrderHandler(
            amr=self.available_amrs.pop(0),
            order=self.order_queue.pop(0),
            server=self,
        )
        self.order_handlers[order_handler.amr.id] = order_handler
        order_handler.handle_next_step()

    def request_next_step(self, amr_id):
        order_handler = self.order_handlers.get(amr_id)
        if order_handler is not None:
            order_handler.handle_next_step()


grid = Grid(
    name="test_grid",
    x_size=1000,
    y_size=1000,
    max_time=5000,
    add_exterior_walls=True,
)

# Create a server
server = Server(grid)

# Create AMRs
amr1 = server.create_amr(
    name="AMR1",
    x_coord=500,
    y_coord=400,
)

amr2 = server.create_amr(
    name="AMR2",
    x_coord=600,
    y_coord=500,
)

server.schedule_order(
    time=5,
    origin_x=450,
    origin_y=450,
    destination_x=500,
    destination_y=500,
)

server.schedule_order(
    time=10,
    origin_x=550,
    origin_y=550,
    destination_x=600,
    destination_y=600,
)

server.schedule_order(
    time=15,
    origin_x=650,
    origin_y=650,
    destination_x=700,
    destination_y=700,
)

# Run the sim
grid.simulate()

from pprint import pp


success = True
for order in server.orders:
    # Uncomment to see the order history
    # print(order.history)
    if order.history[-1]["status"] != "delivered":
        success = False
if success:
    print("test_08.py: passed")
else:
    print("test_08.py: failed")



from fizgrid.grid import Grid
from fizgrid.entities import Entity, StaticEntity
from fizgrid.utils import Shape

# from pprint import pp as print

grid = Grid(
    name="test_grid",
    x_size=20,
    y_size=20,
    add_exterior_walls=True,
)

# Add some AMRs to the grid
amr1 = grid.add_entity(
    Entity(
        name="AMR1",
        shape=[[-1.5, -0.5], [-1.5, 0.5], [1.5, 0.5], [1.5, -0.5]],
        x_coord=5,
        y_coord=5,
        auto_rotate=True,
    )
)

amr2 = grid.add_entity(
    Entity(
        name="AMR2",
        shape=[[-1.5, -0.5], [-1.5, 0.5], [1.5, 0.5], [1.5, -0.5]],
        x_coord=15,
        y_coord=5,
    )
)

wall1 = grid.add_entity(
    StaticEntity(
        name="Wall1",
        shape=[[8, 0], [0, 0], [0, 1], [8, 1]],
        x_coord=11,
        y_coord=10,
    )
)

wall2 = grid.add_entity(
    StaticEntity(
        name="Wall2",
        shape=[[-8, 0], [0, 0], [0, 1], [-8, 1]],
        x_coord=9,
        y_coord=10,
    )
)

amr1.add_route(
    waypoints=[
        (10, 5, 1),
        (10, 15, 1),
        (5, 15, 1),
    ],
)

amr2.add_route(
    waypoints=[
        (10, 5, 1),
        (10, 15, 1),
        (15, 15, 1),
    ],
    time=5,
)

# Run the sim
grid.simulate()

# Check if the auto rotated AMR clears the gap
success = True
try:
    if amr1.history[-1]["y"] != 15:
        success = False
    if amr2.history[-1]["y"] != 9.5:
        success = False
except:
    success = False
if success:
    print("test_11.py: passed")
else:
    print("test_11.py: failed")



from fizgrid.helpers.waypoint_timing.utils import (
    get_distance,
    get_angle,
    get_end_speed,
    partition_distance,
)


def get_max_corner_speeds(
    waypoints: list[tuple], max_speed: float | int, acceleration: int | float
):
    """
    Calculates the maximum speeds that can be taken at each corner in a set of waypoints
    given a constant acceleration / deceleration between 0 and max_speed.

    The first and last waypoints are always 0 speed.

    args:

    - waypoints: A list of tuples representing the waypoints (x, y).
    - max_speed: The maximum speed of the entity.
    - acceleration: The acceleration/deceleration of the entity.
    """
    # Create a list of speeds that offer the max speed for each corner based on the angle
    speeds = []
    for idx, waypoint in enumerate(waypoints):
        if idx == 0 or idx == len(waypoints) - 1:
            speeds.append(0.0)  # Must start and end at 0 speed
        else:
            angle = get_angle(waypoints[idx - 1], waypoint, waypoints[idx + 1])
            # Map tighter angles to lower allowable speeds
            if angle > 178:
                corner_speed = max_speed
            elif angle > 170:
                corner_speed = max_speed * 0.8
            elif angle > 160:
                corner_speed = max_speed * 0.6
            elif angle > 150:
                corner_speed = max_speed * 0.4
            elif angle > 140:
                corner_speed = max_speed * 0.2
            else:
                corner_speed = max_speed * 0.1
            speeds.append(corner_speed)

    # Go backwards to update the maximum speed at each corner based on the deceleration
    # distance needed to not exceed the speed limit on any corner
    for idx in range(
        len(waypoints) - 1
    ):  # Minus 1 to avoid the index out of range error
        distance = get_distance(waypoints[-idx - 1], waypoints[-idx - 2])
        speeds[-idx - 2] = min(
            speeds[-idx - 2],
            get_end_speed(speeds[-idx - 1], distance, acceleration),
        )

    return speeds


def acceleration_waypoint_time_approximation(
    start_x,
    start_y,
    waypoints: list[tuple],
    max_speed: int | float,
    acceleration: int | float,
    round_time_to: int | float = 4,
):
    """
    Partitions a set of waypoints (list of tuples(x,y)) into a list of (x,y,time_delta) tuples
    given a constant acceleration / deceleration between 0 and max_speed.

    Each waypoint can be partitioned into an acceleration, cruise and deceleration phase (as applicable).

    args:

    - start_x: The x coordinate of the starting point.
    - start_y: The y coordinate of the starting point.
    - waypoints: A list of tuples representing the waypoints (x, y).
    - max_speed: The maximum speed of the entity.
    - acceleration: The acceleration/deceleration of the entity.
    - round_time_to: The number of decimal places to round the time to. Default is 4.

    Returns

    - list of tuples:  representing the waypoints (x, y, time_delta).
    """
    partitioned_waypoints = []
    # Add the starting point to the list of waypoints
    waypoints = [(start_x, start_y)] + waypoints
    waypoint_speeds = get_max_corner_speeds(waypoints, max_speed, acceleration)
    current_speed = 0
    for idx in range(len(waypoints) - 1):
        start_point = waypoints[idx]
        end_point = waypoints[idx + 1]

        distance = get_distance(start_point, end_point)
        partitions, current_speed = partition_distance(
            distance=distance,
            start_speed=current_speed,
            max_end_speed=waypoint_speeds[idx + 1],
            max_speed=max_speed,
            acceleration=acceleration,
        )
        for partition in partitions:
            partitioned_waypoints.append(
                (
                    start_point[0]
                    + (end_point[0] - start_point[0]) * partition["end_pct"],
                    start_point[1]
                    + (end_point[1] - start_point[1]) * partition["end_pct"],
                    round(partition["time"], round_time_to),
                )
            )
    return partitioned_waypoints



from fizgrid.grid import Grid
from fizgrid.entities import Entity
from fizgrid.utils import Shape
import random, math

# from pprint import pp as print


class Scheduler:
    def __init__(self, grid):
        self.grid = grid
        self.orders = []
        self.processors = []

    def add_order(self, order):
        self.orders.append(order)
        order.set_status(order.time, "in_queue")
        self.try_assignment()

    def add_processor(self, processor):
        self.processors.append(processor)
        self.try_assignment()

    def try_assignment(self):
        if len(self.orders) == 0 or len(self.processors) == 0:
            return
        order = self.orders.pop(0)
        processor = self.processors.pop(0)
        order.set_status(processor.get_time(), "scheduled")
        processor.add_order(order)


class Order:
    def __init__(
        self,
        time,
        origin_x,
        origin_y,
        destination_x,
        destination_y,
        scheduler: Scheduler,
    ):
        self.time = time
        self.origin_x = origin_x
        self.origin_y = origin_y
        self.destination_x = destination_x
        self.destination_y = destination_y
        self.status = "none"
        self.scheduler = scheduler
        self.history = []

        # Add this order to the event stream
        self.scheduler.grid.add_event(
            time=time,
            object=self,
            method="arrive",
            kwargs={},
        )

    def set_status(self, time, status):
        self.time = time
        self.status = status
        self.history.append({"time": time, "status": status})

    def arrive(self):
        self.set_status(self.time, "order_arrived")
        self.scheduler.add_order(self)


class SnifferAMR(Entity):
    def __init__(self, *args, scheduler: Scheduler, **kwargs):
        super().__init__(*args, **kwargs)
        self.order = None
        self.speed = 1
        self.scheduler = scheduler
        self.scheduler.add_processor(self)

    def add_order(self, order: Order):
        self.order = order
        self.handle_order_state()

    def handle_order_state(self):
        if self.order.status == "scheduled":
            self.set_goal(
                x_coord=self.order.origin_x,
                y_coord=self.order.origin_y,
            )
            self.order.set_status(self.get_time(), "waiting_pickup")
        elif self.order.status == "waiting_pickup":
            self.set_goal(
                x_coord=self.order.destination_x,
                y_coord=self.order.destination_y,
            )
            self.order.set_status(self.get_time(), "picked_up")
        elif self.order.status == "picked_up":
            self.order.set_status(self.get_time(), "delivered")
            self.order = None
            self.scheduler.add_processor(self)

    def get_dist_from_goal(self):
        return (
            (self.goal_x - self.x_coord) ** 2
            + (self.goal_y - self.y_coord) ** 2
        ) ** 0.5

    def set_goal(self, x_coord, y_coord, tolerance=1):
        self.goal_x = x_coord
        self.goal_y = y_coord
        self.tolerance = tolerance
        self.add_next_route()

    def add_next_route(self):
        distance_from_goal = self.get_dist_from_goal()
        if distance_from_goal < self.tolerance:
            self.handle_order_state()
        else:
            goal_angle_rad = math.atan2(
                self.goal_y - self.y_coord, self.goal_x - self.x_coord
            )
            random_angle = random.normalvariate(goal_angle_rad, math.pi / 2)
            distance = random.uniform(0, min(distance_from_goal, 5))
            x_shift = distance * math.cos(random_angle)
            y_shift = distance * math.sin(random_angle)
            self.add_route(
                waypoints=[
                    (
                        self.x_coord + x_shift,
                        self.y_coord + y_shift,
                        distance * self.speed,
                    ),
                ]
            )

    def on_realize(self, **kwargs):
        if self.order:
            self.add_next_route()


grid = Grid(
    name="test_grid",
    x_size=1000,
    y_size=1000,
    max_time=5000,
    add_exterior_walls=True,
)

scheduler = Scheduler(grid)

# Add some AMRs to the grid
amr1 = grid.add_entity(
    SnifferAMR(
        name="AMR1",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=500,
        y_coord=400,
        scheduler=scheduler,
    )
)

amr2 = grid.add_entity(
    SnifferAMR(
        name="AMR2",
        shape=Shape.rectangle(x_len=1, y_len=1, round_to=2),
        x_coord=600,
        y_coord=500,
        scheduler=scheduler,
    )
)

orders = [
    Order(
        time=5,
        origin_x=450,
        origin_y=450,
        destination_x=500,
        destination_y=500,
        scheduler=scheduler,
    ),
    Order(
        time=10,
        origin_x=550,
        origin_y=550,
        destination_x=600,
        destination_y=600,
        scheduler=scheduler,
    ),
    Order(
        time=15,
        origin_x=650,
        origin_y=650,
        destination_x=700,
        destination_y=700,
        scheduler=scheduler,
    ),
]

# Run the sim
next_state_events = True
while next_state_events:
    next_state_events = grid.resolve_next_state()

success = True
for order in orders:
    # Uncomment to see the order history
    # print(order.history)
    if order.history[-1]["status"] != "delivered":
        success = False
        break
if success:
    print("test_06.py: passed")
else:
    print("test_06.py: failed")
